/*
 * GameMode.java
 *
 * This is the primary class file for running the game.  You should study this file for
 * ideas on how to structure your own root class. This class follows a 
 * model-view-controller pattern fairly strictly.
 *
 * Author: Walker M. White
 * Based on original Parallax Sample by Don Holden, 2007
 * LibGDX version, 2/28/2015
 */
package edu.cornell.gdiac.parallax;

import edu.cornell.gdiac.util.*;

import com.badlogic.gdx.*;
import com.badlogic.gdx.assets.*;
import com.badlogic.gdx.graphics.*;

/**
 * The primary controller class for the game.
 *
 * While GDXRoot is the root class, it delegates all of the work to the player mode
 * classes. This is the player mode class for running the game. In initializes all 
 * of the other classes in the game and hooks them together.  It also provides the
 * basic game loop (update-draw).
 */
public class GameMode implements Screen {
	// GRAPHICS AND SOUND RESOURCES
	// Pathnames to texture  and sound assets
	private static String GRASS_FILE   = "parallax/grass.png";
	private static String ROAD_FILE    = "parallax/road.jpg";
	private static String SUNSET_FILE  = "parallax/sunset.jpg";
	private static String MOUNT_FILE   = "parallax/mountains.png";
	private static String WHEEL_FILE   = "parallax/wheel.png";
	private static String CHASSIS_FILE = "parallax/chassis.png";

	// Asset loading is handled statically, so these are static variables
	/** The absolute background layer */
	private static Texture sunsetTexture;
	/** The mountains in front of the sunset */
	private static Texture mountainTexture;
	/** The road underneath our wheels */
	private static Texture roadTexture;
	/** The grass layer */
	private static Texture grassTexture;
	/** The car wheels */
	private static Texture wheelTexture;
	/** The car body */
	private static Texture chassisTexture;
	
	/** 
	 * Preloads the assets for this game.
	 * 
	 * All instance of the game use the same assets, so this is a static method.  
	 * This keeps us from loading the assets multiple times.
	 *
	 * The asset manager for LibGDX is asynchronous.  That means that you
	 * tell it what to load and then wait while it loads them.  This is 
	 * the first step: telling it what to load.
	 * 
	 * @param manager Reference to global asset manager.
	 */
	public static void PreLoadContent(AssetManager manager) {
		// Load the background.
		manager.load(GRASS_FILE,Texture.class);
		manager.load(ROAD_FILE,Texture.class);
		manager.load(SUNSET_FILE,Texture.class);
		manager.load(MOUNT_FILE,Texture.class);
		manager.load(WHEEL_FILE,Texture.class);
		manager.load(CHASSIS_FILE,Texture.class);
	}

	/** 
	 * Loads the assets for this game.
	 * 
	 * All instance of the game use the same assets, so this is a static method.  
	 * This keeps us from loading the assets multiple times.
	 *
	 * The asset manager for LibGDX is asynchronous.  That means that you
	 * tell it what to load and then wait while it loads them.  This is 
	 * the second step: extracting assets from the manager after it has
	 * finished loading them.
	 * 
	 * @param manager Reference to global asset manager.
	 */
	public static void LoadContent(AssetManager manager) {
		// Allocate the textures
		grassTexture = loadTexture(manager,GRASS_FILE);
		roadTexture = loadTexture(manager,ROAD_FILE);
		sunsetTexture = loadTexture(manager,SUNSET_FILE);
		mountainTexture = loadTexture(manager,MOUNT_FILE);
		wheelTexture = loadTexture(manager,WHEEL_FILE);
		chassisTexture = loadTexture(manager,CHASSIS_FILE);
	}
	
	/**
	 * Helper to initialize a texture after loading.
	 * 
	 * @param manager Reference to global asset manager
	 * @param key The key identifying the texture in the loader
	 * 
	 * @return the texture newly initialized
	 */
	private static Texture loadTexture(AssetManager manager, String key) {
		Texture result = null;
		if (manager.isLoaded(key)) {
			result = manager.get(key, Texture.class);
			result.setFilter(Texture.TextureFilter.Linear, Texture.TextureFilter.Linear);
		} 
		return result;
	}

	/** 
	 * Unloads the assets for this game.
	 * 
	 * This method erases the static variables.  It also deletes the associated textures 
	 * from the asset manager.
	 * 
	 * @param manager Reference to global asset manager.
	 */
	public static void UnloadContent(AssetManager manager) {
		if (grassTexture != null) {
			grassTexture = null;
			manager.unload(GRASS_FILE);
		}
		if (roadTexture != null) {
			roadTexture = null;
			manager.unload(ROAD_FILE);
		}
		if (sunsetTexture != null) {
			sunsetTexture = null;
			manager.unload(SUNSET_FILE);
		}
		if (mountainTexture != null) {
			mountainTexture = null;
			manager.unload(MOUNT_FILE);
		}
		if (wheelTexture != null) {
			wheelTexture = null;
			manager.unload(WHEEL_FILE);
		}
		if (chassisTexture != null) {
			chassisTexture = null;
			manager.unload(CHASSIS_FILE);
		}
	}
	
	/** Parallax values */
	private static final float SUNSET_PARALLAX    = 0;  	// Parallax = 0 means we're infinitely far away
	private static final float MOUNTAINS_PARALLAX = 0.05f;
	private static final float BGGRASS_PARALLAX   = 0.3f;
	private static final float ROAD_PARALLAX      = 1.0f;	// Put focus of scene at parallax 1
	private static final float FGGRASS_PARALLAX   = 2.0f;	// Parallax > 1 is a foreground object

	/** Heights at which to draw objects in the game world */
	private static final float SUNSET_HEIGHT    = 0;
	private static final float MOUNTAINS_HEIGHT = 0;
	private static final float BGGRASS_HEIGHT   = 58;
	private static final float FGGRASS_HEIGHT   = -67;
	private static final float ROAD_HEIGHT      = 63;
	private static final float CAR_HEIGHT       = 143;
	private static final float WHEEL_HEIGHT     = 127;

	/** X-offsets */
	private static final float CAR_OFFSET    = 250;
	private static final float WHEEL1_OFFSET = 340;
	private static final float WHEEL2_OFFSET = 457;

	/** How bumpy the ride is */
	private static final float BUMPY_AMPLITUDE = 5.0f;
	private static final float BUMPY_FREQUENCY = 0.02f;
	
	/** Reference to drawing context to display graphics (VIEW CLASS) */
	private GameCanvas canvas;
	
	/** Reads input from keyboard or game pad (CONTROLLER CLASS) */
	private InputController inputController;
	
	/** Whether or not this player mode is still active */
	private boolean active;
	/** Listener that will update the player mode when we are done */
	private ScreenListener listener;

	/** Vehicle state */
    private float carPosition;
    private float wheelRotation;
    private float bumpSine;
    
	/**
	 * Creates a new game with the given drawing context.
	 *
	 * This constructor initializes the models and controllers for the game.  The
	 * view has already been initialized by the root class.
	 */
	public GameMode(GameCanvas canvas) {
		this.canvas = canvas;
		// Create the controllers.
		inputController = new InputController();
	}
	
	/**
	 * Dispose of all (non-static) resources allocated to this mode.
	 */
	public void dispose() {
		inputController = null;
		canvas = null;
	}
	
	
	/**
	 * Update the game state.
	 *
	 * We prefer to separate update and draw from one another as separate methods, instead
	 * of using the single render() method that LibGDX does.  We will talk about why we
	 * prefer this in lecture.
	 *
	 * @param delta Number of seconds since last animation frame
	 */
	private void update(float delta) {
		inputController.readInput();

		// Allows the game to exit
		if (inputController.didExit()) {
			System.exit(0);
		}

        // Change position of car
		carPosition += inputController.getSpeed();
		bumpSine += inputController.getSpeed() * BUMPY_FREQUENCY;

		// Change rotation in degrees
		wheelRotation -= (inputController.getSpeed()*360.0f) / (Math.PI*wheelTexture.getWidth());
		if (wheelRotation > 360.0f) {
			wheelRotation -= 360.0f;
		} else if (wheelRotation < 0.0f) {
			wheelRotation += 360.0f;
		}
	}
	
	/**
	 * Draw the status of this player mode.
	 *
	 * We prefer to separate update and draw from one another as separate methods, instead
	 * of using the single render() method that LibGDX does.  We will talk about why we
	 * prefer this in lecture.
	 */
	private void draw(float delta) {
		canvas.begin();

        // Camera's X-position, 400 pixels (half the screen) to the left of the car.
		float camera = -carPosition;

        // Draw the sunset
		canvas.drawWrapped(sunsetTexture, SUNSET_PARALLAX * camera, SUNSET_HEIGHT);

        // Draw the mountains
		canvas.drawWrapped(mountainTexture, MOUNTAINS_PARALLAX * camera, MOUNTAINS_HEIGHT);

        // Draw the background grass
		canvas.drawWrapped(grassTexture, BGGRASS_PARALLAX * camera, BGGRASS_HEIGHT);

        // Draw the road
		canvas.drawWrapped(roadTexture, ROAD_PARALLAX * camera, ROAD_HEIGHT);

        // Draw the wheels
        float ox = wheelTexture.getWidth()/2.0f;
        float oy = wheelTexture.getHeight()/2.0f;
		canvas.draw(wheelTexture, Color.WHITE, ox, oy, ROAD_PARALLAX * (WHEEL1_OFFSET), WHEEL_HEIGHT, 
					wheelRotation, 1.0f, 1.0f);
        canvas.draw(wheelTexture, Color.WHITE, ox, oy, ROAD_PARALLAX * (WHEEL2_OFFSET), WHEEL_HEIGHT, 
					wheelRotation, 1.0f, 1.0f);

        // Draw the car
        ox = chassisTexture.getWidth()/2.0f;
        oy = chassisTexture.getHeight()/2.0f;
        float bump = BUMPY_AMPLITUDE * (float)Math.sin(bumpSine);
		canvas.draw(chassisTexture, Color.WHITE, 0.0f, 0.0f, ROAD_PARALLAX * (CAR_OFFSET), CAR_HEIGHT + bump, 
					0.0f, 1.0f, 1.0f);
					
        // Draw the foreground grass
		canvas.drawWrapped(grassTexture, FGGRASS_PARALLAX * camera, FGGRASS_HEIGHT);

		canvas.end();		
	}
	
	/**
	 * Called when the Screen is resized. 
	 *
	 * This can happen at any point during a non-paused state but will never happen 
	 * before a call to show().
	 *
	 * @param width  The new width in pixels
	 * @param height The new height in pixels
	 */
	public void resize(int width, int height) {
		// IGNORE FOR NOW
	}

	/**
	 * Called when the Screen should render itself.
	 *
	 * We defer to the other methods update() and draw().  However, it is VERY important
	 * that we only quit AFTER a draw.
	 *
	 * @param delta Number of seconds since last animation frame
	 */
	public void render(float delta) {
		if (active) {
			update(delta);
			draw(delta);
			if (inputController.didExit() && listener != null) {
				listener.exitScreen(this, 0);
			}
		}
	}

	/**
	 * Called when the Screen is paused.
	 * 
	 * This is usually when it's not active or visible on screen. An Application is 
	 * also paused before it is destroyed.
	 */
	public void pause() {
		// TODO Auto-generated method stub
	}

	/**
	 * Called when the Screen is resumed from a paused state.
	 *
	 * This is usually when it regains focus.
	 */
	public void resume() {
		// TODO Auto-generated method stub
	}
	
	/**
	 * Called when this screen becomes the current screen for a Game.
	 */
	public void show() {
		// Useless if called in outside animation loop
		active = true;
	}

	/**
	 * Called when this screen is no longer the current screen for a Game.
	 */
	public void hide() {
		// Useless if called in outside animation loop
		active = false;
	}

	/**
	 * Sets the ScreenListener for this mode
	 *
	 * The ScreenListener will respond to requests to quit.
	 */
	public void setScreenListener(ScreenListener listener) {
		this.listener = listener;
	}
}
